#version 430

in vec2 uv;
in vec3 fragNormal;
in vec3 fragPos;
in vec4 fragPosLight;
in vec4 fragPosLight2;

uniform vec3 eyePos;
uniform sampler2D texture2d;
uniform vec4 materialCoefficients;
uniform sampler2DShadow depthMap;
uniform sampler2DShadow depthMap2;

uniform float brightness;

const int MAX_DIR_LIGHTS = 1;
const int MAX_SPOT_LIGHTS = 2;

struct DirLight {
    vec3 direction;
	vec3 color;
};  
uniform DirLight dirLight[MAX_DIR_LIGHTS];
// spotlight struct
struct SpotLight {
    vec3 direction;
	vec3 color;
	vec3 position;
	vec3 attenuation;
	float cutOff;
	float outerCutOff;
};  
uniform SpotLight spotLight[MAX_SPOT_LIGHTS];

out vec4 fragColor;

vec3 phong(vec3 n, vec3 l, vec3 v, vec3 diffuseC, float diffuseF, vec3 specularC, float specularF, float alpha);
float ShadowCalculation(vec4 fragPosLightSpace, float bias, sampler2DShadow dM );

void main(){

		// re-normalize fragment normals after interpolation
		vec3 normal = normalize(fragNormal);

		// re-normalize view direction
		vec3 viewDir = normalize(eyePos - fragPos);

		// texture
		vec3 texColor = texture(texture2d, uv).rgb;

		// ambient 
		fragColor = vec4(texColor * materialCoefficients.x * brightness, 1);
		
		// directional lights
		vec3 directionallight = vec3(0.0f);
		for (int k = 0; k < MAX_DIR_LIGHTS; k++){
			vec3 lightDir = normalize(-dirLight[k].direction);
			vec3 reflectDir = reflect(-lightDir, normal);
			vec3 diffuse = dirLight[k].color*texColor * materialCoefficients.y * max(0, dot(normal, lightDir));
			vec3 specular = dirLight[k].color * materialCoefficients.z * pow(max(0, dot(reflectDir, viewDir)), materialCoefficients.w); 
						
			// calculate shadow
			float bias = 0.0f;//max(0.005 * (1.0 - dot(normal, lightDir)), 0.0005);  
			float shadow = ShadowCalculation(fragPosLight, bias, depthMap);

			directionallight += (diffuse + specular) *(1.0-shadow);
		}

		// spotLight
		vec3 spot = vec3(0.0f);
		for (int l = 0; l < MAX_SPOT_LIGHTS; l++){

			// soft edges
			vec3 lightDir = normalize(spotLight[l].position - fragPos);
			float theta = dot(lightDir, normalize(-spotLight[l].direction));   
			float epsilon = spotLight[l].cutOff - spotLight[l].outerCutOff;
			float intensity = clamp((theta - spotLight[l].outerCutOff) / epsilon, 0.0, 1.0);
			
			if (theta > spotLight[l].outerCutOff){

				// attenuation
				float dist = length(spotLight[l].position - fragPos);
				float att = 1.0 / (spotLight[l].attenuation.x + spotLight[l].attenuation.y * dist + spotLight[l].attenuation.z * (dist * dist)); 
			
				// diffuse 
				float diff = max(dot(normal, lightDir), 0.0);
				vec3 diffuse = diff * spotLight[l].color * materialCoefficients.y * texColor;  
				diffuse *= att;
				diffuse *= intensity;
        
				// specular
				vec3 reflectDir = normalize(reflect(-lightDir, normal));  
				float spec = pow(max(dot(viewDir, reflectDir), 0.0), materialCoefficients.w);
				vec3 specular = spec * spotLight[l].color * materialCoefficients.z;  				
				specular *= att;   
				specular *= intensity;
								
				// add everything
				vec3 result = diffuse + specular;
				
				// calculate shadow
				if (l == 1){
					float bias = 0.0;//max(0.005 * (1.0 - dot(normal, lightDir)), 0.0005);  
					float shadow = ShadowCalculation(fragPosLight2, bias, depthMap2);            
					result *= (1.0-shadow);				
				}
				
				spot += result;
			}
		}
	
		// add directional light contribution
		fragColor.rgb += directionallight;	
		fragColor.rgb += spot;
		
}

float ShadowCalculation(vec4 fragPosLightSpace, float bias, sampler2DShadow dM)
{
	
	float shadow = 0.0;
	// perform perspective divide in [-1,1] space (important for perspective projection)
    vec3 projectedCoord = fragPosLightSpace.xyz / fragPosLightSpace.w;

	// to avoid shadow beyond the far plane
	if(projectedCoord.z > 1.0){
		return shadow;
	}

	// transform to [0,1]
	projectedCoord = 0.5* projectedCoord + 0.5;	
	projectedCoord.z -= bias;
	// 4 depth comparison are made -> return value in [0,1]
	shadow = texture(dM , projectedCoord);
	shadow = (1.0-shadow);

	return shadow;
}