#version 430

// inputs
in vec2 uv;
in vec3 fragNormal;
in vec3 fragPos;
in vec4 fragPosLight;
in vec4 fragPosLight2;

// outputs
out vec4 fragColor;

// camera position
uniform vec3 eyePos;

// depth maps
uniform sampler2DShadow depthMap;
uniform sampler2DShadow depthMap2;

// material parameters 
uniform vec3  albedo;
uniform float metallic;
uniform float roughness;
uniform float ao;

uniform float brightness;

const float PI = 3.14159265359;
const int MAX_DIR_LIGHTS = 1;
const int MAX_SPOT_LIGHTS = 2;

// directional light struct
struct DirLight {
    vec3 direction;
	vec3 color;
};  
uniform DirLight dirLight[MAX_DIR_LIGHTS];
// spotlight struct
struct SpotLight {
    vec3 direction;
	vec3 color;
	vec3 position;
	vec3 attenuation;
	float cutOff;
	float outerCutOff;
};  
uniform SpotLight spotLight[MAX_SPOT_LIGHTS];


// methods
float DistributionGGX(vec3 N, vec3 H, float roughness);
float GeometrySchlickGGX(float NdotV, float roughness);
float GeometrySmith(vec3 N, vec3 V, vec3 L, float roughness);
vec3 fresnelSchlick(float cosTheta, vec3 F0);
float ShadowCalculation(vec4 fragPosLightSpace, float bias, sampler2DShadow dM);

void main(){

		// re-normalize fragment normals after interpolation
		vec3 normal = normalize(fragNormal);

		// re-normalize view direction
		vec3 viewDir = normalize(eyePos - fragPos);
		
		// F0
		vec3 F0 = vec3(0.04); 
		F0 = mix(F0, albedo, metallic);

		// shadow
		float shadow = 0.0;

		// reflectance equation
		vec3 Lo = vec3(0.0);

		// directional light
		for (int k = 0; k < MAX_DIR_LIGHTS; k++){

			// calculate per-light radiance
			vec3 lightDir = normalize(-dirLight[k].direction);
			vec3 halfwayVec = normalize(viewDir + lightDir);
			vec3 radiance = dirLight[k].color;        
        
			// cook-torrance brdf
			float D = DistributionGGX(normal, halfwayVec, roughness);        
			float G = GeometrySmith(normal, viewDir, lightDir, roughness);      
			vec3 F = fresnelSchlick(max(dot(halfwayVec, viewDir), 0.0), F0);       
        
			vec3 kS = F;
			vec3 kD = vec3(1.0) - kS;
			kD *= 1.0 - metallic;	  
        
			vec3 num = D * G * F;
			float denom = 4.0 * max(dot(normal, viewDir), 0.0) * max(dot(normal, lightDir), 0.0);
			vec3 specular = num / max(denom, 0.001);  

			// calculate shadow
			float bias =0.0;// max(0.005 * (1.0 - dot(normal, lightDir)), 0.0005);  
			shadow += ShadowCalculation(fragPosLight, bias, depthMap);
            
			// add to outgoing radiance Lo
			float NdotL = max(dot(normal, lightDir), 0.0);                
			Lo += (kD * albedo / PI + specular) * radiance * NdotL * (1.0 - shadow); 

		}
		// spotlight
		for (int l = 0; l < MAX_SPOT_LIGHTS; l++){

			// soft edges
			vec3 lightDir = normalize(spotLight[l].position - fragPos);
			float theta = dot(lightDir, normalize(-spotLight[l].direction));   
			float epsilon = spotLight[l].cutOff - spotLight[l].outerCutOff;
			float intensity = clamp((theta - spotLight[l].outerCutOff) / epsilon, 0.0, 1.0);

			if (theta > spotLight[l].outerCutOff){

				// calculate per-light radiance			
				vec3 halfwayVec = normalize(viewDir + lightDir);
				float dist = length(spotLight[l].position - fragPos);
				float attenuation = 1.0 / (spotLight[l].attenuation.x + spotLight[l].attenuation.y * dist + spotLight[l].attenuation.z * (dist * dist)); 
				vec3 radiance = spotLight[l].color * attenuation * intensity;  
        
				// cook-torrance brdf
				float NDF = DistributionGGX(normal, halfwayVec, roughness);        
				float G = GeometrySmith(normal, viewDir, lightDir, roughness);      
				vec3 F = fresnelSchlick(max(dot(normal, viewDir), 0.0), F0); 
        
				vec3 kS = F;
				vec3 kD = vec3(1.0) - kS;
				kD *= 1.0 - metallic;	  
        
				vec3 num = NDF * G * F;
				float denom = 4.0 * max(dot(normal, viewDir), 0.0) * max(dot(normal, lightDir), 0.0);
				vec3 specular = num / max(denom, 0.001);  
            
				// add to outgoing radiance Lo
				float NdotL = max(dot(normal, lightDir), 0.0);   
				vec3 result = (kD * albedo / PI +  specular ) *  radiance * NdotL; 

				// calculate shadow
				if (l == 1){
					float bias = 0.0;//max(0.005 * (1.0 - dot(normal, lightDir)), 0.0005);  
					float shadow = ShadowCalculation(fragPosLight2, bias, depthMap2);            
					result *= (1.0-shadow);				
				}
				
				Lo += result;
				
			}
		}
		
		// ambient light
		vec3 ambient = vec3(0.03) * albedo * ao * brightness;

		// resulting light
		vec3 color = (ambient + Lo) ;

		// HDR and gama correction
		color = color / (color + vec3(1.0));
		color = pow(color, vec3(1.0/2.2));  
   
		// fragment color
		fragColor = vec4(color, 1.0);		
}


float DistributionGGX(vec3 normal, vec3 halfway, float roughness)
{
	float a      = roughness*roughness;
	float a2     = a*a;
	float NdotH  = max(dot(normal, halfway), 0.0);
	float NdotH2 = NdotH*NdotH;
	
	float num   = a2;
	float denom = (NdotH2 * (a2 - 1.0) + 1.0);
	denom = PI * denom * denom;
	
	return num / denom;
}

float GeometrySchlickGGX(float NdotV, float roughness)
{
	float r = (roughness + 1.0);
	float k = (r*r) / 8.0;

	float num   = NdotV;
	float denom = NdotV * (1.0 - k) + k;
	
	return num / denom;
}

float GeometrySmith(vec3 normal, vec3 view, vec3 lightDir, float roughness)
{
	float NdotV = max(dot(normal, view), 0.0);
	float NdotL = max(dot(normal, lightDir), 0.0);
	float ggx2  = GeometrySchlickGGX(NdotV, roughness);
	float ggx1  = GeometrySchlickGGX(NdotL, roughness);
	
	return ggx1 * ggx2;
}

vec3 fresnelSchlick(float cosTheta, vec3 F0)
{
	vec3 fresnel = F0 + (1.0 - F0) * pow(1.0 - cosTheta, 5.0);
    return max(fresnel, 0.0);
}

float ShadowCalculation(vec4 fragPosLightSpace, float bias, sampler2DShadow dM)
{
	// perform perspective divide in [-1,1] space (important for perspective projection)
    vec3 projectedCoord = fragPosLightSpace.xyz / fragPosLightSpace.w;

	float shadow = 0.0;
	// to avoid shadow beyond the far plane
	if(projectedCoord.z > 1.0){
		return shadow;
	}

	// transform to [0,1]
	projectedCoord = 0.5* projectedCoord + 0.5;	
	// 4 depth comparison are made -> return value in [0,1]
	shadow = texture(dM , projectedCoord);
	shadow = (1.0-shadow);

	return shadow;
}
