#version 430

in vec2 uv;
in vec3 fragNormal;
in vec3 fragPos;
in vec4 fragPosLight;
in vec4 fragPosLight2;

uniform vec3 eyePos;
uniform sampler2DShadow depthMap;
uniform sampler2DShadow depthMap2;

uniform float brightness;

const int MAX_DIR_LIGHTS = 1;
const int MAX_SPOT_LIGHTS = 2;

struct DirLight {
    vec3 direction;
	vec3 color;
	vec3 coefficients;
};  
uniform DirLight dirLight[MAX_DIR_LIGHTS];
struct SpotLight {
    vec3 direction;
	vec3 color;
	vec3 position;
	vec3 attenuation;
	float cutOff;
	float outerCutOff;
	vec3 coefficients;
};  
uniform SpotLight spotLight[MAX_SPOT_LIGHTS];
struct Material {
    sampler2D diffuse;
    sampler2D specular;
    float shininess;
}; 
uniform Material material;

out vec4 fragColor;

float ShadowCalculation(vec4 fragPosLightSpace, float bias, sampler2DShadow dM);

void main(){

		// re-normalize fragment normals after interpolation
		vec3 normal = normalize(fragNormal);

		// re-normalize view direction
		vec3 viewDir = normalize(eyePos - fragPos);

		// texture
		vec3 texColor = texture(material.diffuse, uv).rgb;
		vec3 texSpecular = texture(material.specular, uv).rgb;

		vec3 ambient = 0.2f * texColor * brightness;
		fragColor = vec4(ambient, 1.0f);
		
		// directional lights
		vec4 directionallight = vec4(0.0f, 0.0f, 0.0f, 0.0f);
		for (int k = 0; k < 1; k++){
  	
			// diffuse 
			vec3 lightDir = normalize(-dirLight[k].direction);
			float diff = max(dot(normal, lightDir), 0.0);
			vec3 diffuse = dirLight[k].coefficients.y * dirLight[k].color * (diff * texColor);
    
			// specular
			vec3 reflectDir = reflect(-lightDir, normal);  
			float spec = pow(max(dot(viewDir, reflectDir), 0.0), material.shininess);
			vec3 specular = dirLight[k].coefficients.z * dirLight[k].color * (spec * texSpecular);  
        
			// calculate shadow
			float bias = 0.0005f;//max(0.005 * (1.0 - dot(normal, lightDir)), 0.0005);  
			float shadow = ShadowCalculation(fragPosLight, bias, depthMap);
			vec3 result = (diffuse + specular)*(1.0-shadow);
			directionallight += vec4(result, 0.0);
		}	

		// spotlights
		vec4 spotlight = vec4(0.0f, 0.0f, 0.0f, 0.0f);
		for (int l = 0; l < MAX_SPOT_LIGHTS; l++) {

			// soft edges
			vec3 lightDir = normalize(spotLight[l].position - fragPos);
			float theta = dot(lightDir, normalize(-spotLight[l].direction));   
			float epsilon = spotLight[l].cutOff - spotLight[l].outerCutOff;
			float intensity = clamp((theta - spotLight[l].outerCutOff) / epsilon, 0.0, 1.0);

			if (theta > spotLight[l].outerCutOff){
				// attenuation
				float dist = length(spotLight[l].position - fragPos);
				float att = 1.0 / (spotLight[l].attenuation.x + spotLight[l].attenuation.y * dist + spotLight[l].attenuation.z * (dist * dist)); 
        
				// diffuse 
				float diff = max(dot(normal, lightDir), 0.0);
				vec3 diffuse = spotLight[l].coefficients.y * diff * spotLight[l].color * texColor;  
				diffuse *= att;
				diffuse *= intensity;
        
				// specular
				vec3 reflectDir = reflect(-lightDir, normal);  
				float spec = pow(max(dot(viewDir, reflectDir), 0.0), material.shininess);
				vec3 specular = spotLight[l].coefficients.z * spec * spotLight[l].color * texSpecular;  				
				specular *= att;   
				specular *= intensity;
            
				// add everything
				vec3 result = diffuse + specular;
				
				// calculate shadow
				if (l == 1){
					float bias = 0.01f;//max(0.005 * (1.0 - dot(normal, lightDir)), 0.0005);  
					float shadow = ShadowCalculation(fragPosLight2, bias, depthMap2);            
					result *= (1.0-shadow);				
				}
				
				spotlight += vec4(result, 0.0f);
			}
		}

		// add directional light and spotlight contribution
		fragColor += directionallight;	
		fragColor += spotlight;
}

float ShadowCalculation(vec4 fragPosLightSpace, float bias, sampler2DShadow dM)
{
	
	float shadow = 0.0;
	// perform perspective divide in [-1,1] space (important for perspective projection)
    vec3 projectedCoord = fragPosLightSpace.xyz / fragPosLightSpace.w;

	// to avoid shadow beyond the far plane
	if(projectedCoord.z > 1.0){
		return shadow;
	}

	// transform to [0,1]
	projectedCoord = 0.5* projectedCoord + 0.5;	
	projectedCoord.z -= bias;
	// 4 depth comparison are made -> return value in [0,1]
	shadow = texture(dM , projectedCoord);
	shadow = (1.0-shadow);

	return shadow;
}